use std::cell::RefCell;
use std::rc::Rc;

use crate::math::{Point2, Region, Region1, Region2, Scalar};

#[derive(Clone, Copy, Debug)]
pub struct Var<T: Clone, TRegion: Region<T>> {
    value: T,
    constraints: TRegion,
}

impl<T: Clone, TRegion: Region<T>> Var<T, TRegion> {
    pub fn new(value: T, constraints: TRegion) -> Self {
        Self { value, constraints }
    }

    pub fn new_full(value: T) -> Self {
        Self::new(value, TRegion::full())
    }

    pub fn new_single(value: T) -> Self {
        Self::new(value.clone(), TRegion::singleton(value))
    }

    pub fn constraints(&self) -> &TRegion {
        &self.constraints
    }

    pub fn reconstrain(&mut self, new_constraints: TRegion) -> bool {
        self.constraints = new_constraints;
        if let Some(n) = self.constraints.nearest(&self.value) {
            self.value = n;
            true
        } else {
            false
        }
    }
}

type ScalarVar = Var<Scalar, Region1>;
type PointVar = Var<Point2, Region2>;

#[derive(Debug)]
pub struct Point {
    // pub id: i64,
    pub pos: PointVar,
}

pub type PointRef = Rc<RefCell<Point>>;

impl Point {
    pub fn new_ref(pos: PointVar) -> PointRef {
        Rc::new(RefCell::new(Point { pos }))
    }
}

struct Line {
    p1: PointRef,
    p2: PointRef,
    len: ScalarVar,
    dir: ScalarVar,
}

// struct System {
//     points: Vec<Point>,
// }
